#pragma once

#define KERNEL_OFFSET 0xFFFFFF8000000000
#define KERNEL_BRK0   0xFFFFFFC000000000

#define PAGE_PRESENT 0x001
#define PAGE_WRITE 0x002

#define PAGE_SIZE 0x1000
#define ENTRIES_PER_PT 512

#ifdef __ASSEMBLER__
    #define V2P(a) ((a) - KERNEL_OFFSET)
    #define P2V(a) ((a) + KERNEL_OFFSET)
#else
    #include <stdint.h>
    #include <stddef.h>
    #define V2P(a) ((uintptr_t)(a) &~KERNEL_OFFSET)
    #define P2V(a) ((void *)((uintptr_t)(a) | KERNEL_OFFSET))

    #define incptr(p, n) ((void *)(((uintptr_t)(p)) + (n)))

    #define P1_OFFSET(a) (((a)>>12) & 0x1FF)
    #define P2_OFFSET(a) (((a)>>21) & 0x1FF)
    #define P3_OFFSET(a) (((a)>>30) & 0x1FF)
    #define P4_OFFSET(a) (((a)>>39) & 0x1FF)

    // memory/memory.c
    void memory_init();
    extern uint64_t kernel_P4;

    // memory/vmm.c
    void vmm_init();
    uint64_t vmm_get_page(uint64_t P4, uint64_t addr);
    int vmm_set_page(uint64_t P4, uint64_t addr, uint64_t page, uint16_t flags);
    void vmm_clear_page(uint64_t P4, uint64_t addr, int free);

    // memory/pmm.c
    void pmm_free(uint64_t page);
    uint64_t pmm_alloc();
    uint64_t pmm_calloc();

    // Link.ld
    extern int kernel_start, kernel_end;

    // boot/boot_PT.s
    extern uint64_t BootP4;

#endif