#include <stdint.h>
#include <proc.h>
#include <memory.h>

struct swtch_stack
{
    uint64_t RBP;
    uint64_t RBX;
    uint64_t R12;
    uint64_t R13;
    uint64_t R14;
    uint64_t R15;
    uint64_t RBP2;
    uint64_t ret;
};

static uint64_t next_tid = 1;

static struct thread *run_queue = 0;
static struct thread *current_thread;
static struct thread *scheduler_thread;

struct thread *new_thread(void (*function)(void))
{
    struct thread *th = P2V(pmm_calloc());
    th->tid = next_tid++;
    th->stack = incptr(th, PAGE_SIZE - sizeof(struct swtch_stack));
    th->q_next = 0;

    struct swtch_stack *stck = th->stack;
    stck->RBP = (uint64_t)&stck->RBP2;
    stck->ret = (uint64_t)function;

    return th;
}

struct thread *thread()
{
    return current_thread;
}

void k_yield()
{
    switch_stack(&current_thread->stack, &scheduler_thread->stack);
}

struct thread *scheduler_next()
{
    if(!current_thread) return run_queue;
    return current_thread->q_next;
}

void scheduler()
{
    while(1)
    {
        struct thread *new = 0;
        while(!(new = scheduler_next()));

        current_thread = new;
        switch_stack(&scheduler_thread->stack, &new->stack);

        current_thread = 0;
    }
}

void start_scheduler()
{
    scheduler_thread = new_thread(scheduler);
    scheduler_thread->tid = (uint64_t)-1;

    uint64_t stack;
    switch_stack(&stack, &scheduler_thread->stack);
}